import numpy as np
import casadi as ca
from scipy.optimize import least_squares

from hurodes.hrdf.joint_mapping.base_solver import BaseSolver

def euler_to_rotmat(roll, pitch, yaw):
    """
    Convert Euler angles (ZYX, yaw-pitch-roll) to rotation matrix
    Input: roll, pitch, yaw (casadi.SX or casadi.MX or float)
    Output: 3x3 rotation matrix (casadi.SX)
    """
    cr = ca.cos(roll)
    sr = ca.sin(roll)
    cp = ca.cos(pitch)
    sp = ca.sin(pitch)
    cy = ca.cos(yaw)
    sy = ca.sin(yaw)

    R = ca.vertcat(
        ca.horzcat(cy*cp, cy*sp*sr - sy*cr, cy*sp*cr + sy*sr),
        ca.horzcat(sy*cp, sy*sp*sr + cy*cr, sy*sp*cr - cy*sr),
        ca.horzcat(-sp,   cp*sr,             cp*cr)
    )
    return R

def get_phi(p_u, p_a, h, r):
    d_ly = p_a[1] - p_u[1]
    l_xz = ca.sqrt(h**2 - d_ly**2)

    delta_x = ca.fabs(p_a[0] - p_u[0])
    delta_z = p_a[2] - p_u[2]
    delta_l = ca.sqrt(delta_x**2 + delta_z**2)

    alpha = ca.arctan2(delta_x, delta_z)
    val = (delta_l**2 + r**2 - l_xz**2) / (2 * r * delta_l)
    val = ca.fmax(ca.fmin(val, 1.0), -1.0)  # Clamp val to [-1, 1]
    beta = ca.acos(val)
    phi = alpha + beta - ca.pi / 2

    return -phi * ca.sign(p_u[0])

class DoubleLinkSolver(BaseSolver):
    def __init__(self, solver_params: dict):
        super().__init__(solver_params)
        self.jacobian_func = self._generate_jacobian_func()

    def inverse(self, pitch, roll):
        p_lu_3 = ca.vertcat(self.solver_params["u_x"], +self.solver_params["d1"], self.solver_params["u_z"])
        p_ru_3 = ca.vertcat(self.solver_params["u_x"], -self.solver_params["d2"], self.solver_params["u_z"])
        p_la_1 = ca.vertcat(0, +self.solver_params["d1"], self.solver_params["h1"])
        p_ra_1 = ca.vertcat(0, -self.solver_params["d2"], self.solver_params["h2"])
        
        p_lu_1 = euler_to_rotmat(roll, pitch, 0) @ p_lu_3
        p_ru_1 = euler_to_rotmat(roll, pitch, 0) @ p_ru_3
        phi_l = get_phi(p_lu_1, p_la_1, self.solver_params["h1"], self.solver_params["r1"])
        phi_r = get_phi(p_ru_1, p_ra_1, self.solver_params["h2"], self.solver_params["r2"])
        return phi_l, phi_r

    def joint2motor_pos(self, joint_pos: np.ndarray):
        """
        Position mapping: motor_pos = f(joint_pos)
        where f is the inverse function of the joint2motor_pos function
        
        Args:
            joint_pos: Current joint positions
        """
        assert len(joint_pos) == 2, f"Joint position must have 2 elements: {len(joint_pos)}"
        pitch, roll = joint_pos
        phi_l, phi_r = self.inverse(pitch, roll)
        motor_pos = np.array([float(phi_l), float(phi_r)])
        return motor_pos

    def _generate_jacobian_func(self):
        roll = ca.SX.sym('roll')
        pitch = ca.SX.sym('pitch')
        phi_l, phi_r = self.inverse(pitch, roll)
        jac = ca.jacobian(ca.vertcat(phi_l, phi_r), ca.vertcat(pitch, roll))
        func = ca.Function('jacobian', [pitch, roll], [jac])
        return func

    def motor2joint_pos(self, motor_pos: np.ndarray, initial_guess: np.ndarray = None):
        """
        Forward kinematics: compute joint positions from motor positions
        Using numerical optimization
        
        Args:
            motor_pos: Motor positions
            initial_guess: Initial guess for optimization, defaults to zero vector if not provided
        """        
        if initial_guess is None:
            initial_guess = np.zeros(2)
        
        def residual(joint_pos):
            computed_motor_pos = self.joint2motor_pos(joint_pos)
            return computed_motor_pos - motor_pos
        
        result = least_squares(residual, initial_guess, method='lm')
        joint_pos = result.x
        return joint_pos

    def joint2motor_vel(self, joint_pos: np.ndarray, joint_vel: np.ndarray):
        """
        Velocity mapping: motor_vel = J * joint_vel
        where J is the Jacobian matrix d(motor_pos)/d(joint_pos)
        
        Args:
            joint_pos: Current joint positions (used to compute Jacobian)
            joint_vel: Joint velocities
        """
        pitch, roll = joint_pos
        J = np.array(self.jacobian_func(pitch, roll))
        motor_vel = J @ joint_vel
        return motor_vel

    def motor2joint_vel(self, joint_pos: np.ndarray, motor_vel: np.ndarray):
        """
        Inverse velocity mapping: joint_vel = J^{-1} * motor_vel
        
        Args:
            joint_pos: Current joint positions (used to compute Jacobian)
            motor_vel: Motor velocities
        """
        pitch, roll = joint_pos
        J = np.array(self.jacobian_func(pitch, roll))
        joint_vel = np.linalg.solve(J, motor_vel)
        return joint_vel

    def joint2motor_torque(self, joint_pos: np.ndarray, joint_torque: np.ndarray):
        """
        Torque mapping: motor_torque = J^{-T} * joint_torque
        
        Args:
            joint_pos: Current joint positions (used to compute Jacobian)
            joint_torque: Joint torques
        """
        pitch, roll = joint_pos
        J = np.array(self.jacobian_func(pitch, roll))
        motor_torque = np.linalg.solve(J.T, joint_torque)
        return motor_torque

    def motor2joint_torque(self, joint_pos: np.ndarray, motor_torque: np.ndarray):
        """
        Inverse torque mapping: joint_torque = J^T * motor_torque
        
        Args:
            joint_pos: Current joint positions (used to compute Jacobian)
            motor_torque: Motor torques
        """
        pitch, roll = joint_pos
        J = np.array(self.jacobian_func(pitch, roll))
        joint_torque = J.T @ motor_torque
        return joint_torque


if __name__ == "__main__":
    solver_params = {
        "d1": 0.035 / 2.0,
        "d2": 0.035 / 2.0,
        "h1": 0.10,
        "h2": 0.17,
        "r1": 0.04,
        "r2": 0.04,
        "u_x": -0.0445,
        "u_z": 0.00,
    }
    
    solver = DoubleLinkSolver(solver_params)

    # Test position mapping
    print("=" * 60)
    print("Test Position Mapping")
    joint_pos = np.array([0.1, 0.05])
    print(f"Joint position (pitch, roll): {joint_pos}")
    motor_pos = solver.joint2motor_pos(joint_pos)
    print(f"Motor position (phi_l, phi_r): {motor_pos}")
    
    # Test forward kinematics
    recovered_joint_pos = solver.motor2joint_pos(motor_pos)
    print(f"Recovered joint position: {recovered_joint_pos}")
    print(f"Error: {np.linalg.norm(joint_pos - recovered_joint_pos)}")
    
    # Test velocity mapping
    print("\n" + "=" * 60)
    print("Test Velocity Mapping")
    joint_vel = np.array([1.0, 0.5])
    print(f"Joint velocity: {joint_vel}")
    motor_vel = solver.joint2motor_vel(joint_pos, joint_vel)
    print(f"Motor velocity: {motor_vel}")
    recovered_joint_vel = solver.motor2joint_vel(joint_pos, motor_vel)
    print(f"Recovered joint velocity: {recovered_joint_vel}")
    print(f"Error: {np.linalg.norm(joint_vel - recovered_joint_vel)}")
    
    # Test torque mapping
    print("\n" + "=" * 60)
    print("Test Torque Mapping")
    joint_torque = np.array([2.0, 1.0])
    print(f"Joint torque: {joint_torque}")
    motor_torque = solver.joint2motor_torque(joint_pos, joint_torque)
    print(f"Motor torque: {motor_torque}")
    recovered_joint_torque = solver.motor2joint_torque(joint_pos, motor_torque)
    print(f"Recovered joint torque: {recovered_joint_torque}")
    print(f"Error: {np.linalg.norm(joint_torque - recovered_joint_torque)}")
    
    # Verify virtual work principle
    print("\n" + "=" * 60)
    print("Verify Virtual Work Principle (should be equal)")
    print(f"Joint virtual work: {np.dot(joint_torque, joint_vel)}")
    print(f"Motor virtual work: {np.dot(motor_torque, motor_vel)}")
    print(f"Virtual work difference: {abs(np.dot(joint_torque, joint_vel) - np.dot(motor_torque, motor_vel))}")



from pathlib import Path
from dataclasses import dataclass
from typing import Optional
from functools import partial

import yaml

from hurodes.hrdf.base.info import InfoBase, InfoList
from hurodes.hrdf.infos import INFO_CLASS_DICT
from hurodes.utils.mesh import simplify_obj
from hurodes import ROBOTS_PATH
from hurodes.utils.config import BaseConfig
from hurodes.hrdf.joint_mapping.joint_mapping_config import JointMappingConfig


class GroundConfig(BaseConfig):
    type: str = "plane"
    contact_affinity: int = 1
    contact_type: int = 1
    friction: float = 1.0

class SimulatorConfig(BaseConfig):
    timestep: float = 0.002
    gravity: list[float] = [0, 0, -9.81]
    ground: GroundConfig = GroundConfig()

class BodyNameConfig(BaseConfig):
    torso_name: str = ""
    hip_names: list[str] = []
    knee_names: list[str] = []
    foot_names: list[str] = []

class DeviceConfig(BaseConfig):
    device_type: str = ""
    device_config_name: str = ""
    name: str = ""

    def model_post_init(self, __context: any):
        self.name = f"{self.device_type}_{self.device_config_name}"

class IMUConfig(DeviceConfig):
    position: list[float] = [0, 0, 0]
    orientation: list[float] = [1, 0, 0, 0]
    body_name: str = ""
    value: list[str] = []

class MotorConfig(DeviceConfig):
    pass

class HRDF:
    def __init__(self, info_class_dict: Optional[dict[str, type[InfoBase]]] = None, **kwargs):
        self.info_class_dict = info_class_dict or INFO_CLASS_DICT
        self.info_list_dict = {name: InfoList(info_class) for name, info_class in self.info_class_dict.items()}

        self.robot_name = None
        self.body_parent_id: list[int] = []
        self.mesh_file_type = None
        self.simulator_config = None
        self.body_name_config = BodyNameConfig()
        self.joint_mapping_config = JointMappingConfig()
        self.imu_config_list = []
        self.motor_config_list = []
    
    @classmethod
    def from_dir(cls, hrdf_path: Path):
        assert hrdf_path.exists(), f"HRDF path not found: {hrdf_path}"
        assert hrdf_path.is_dir(), f"HRDF path is not a directory: {hrdf_path}"
        assert (hrdf_path / "meta.yaml").exists(), f"meta.yaml not found in HRDF path: {hrdf_path}"

        instance = cls()
        with open(Path(hrdf_path, "meta.yaml"), "r", encoding='utf-8') as f:
            meta_info = yaml.safe_load(f)
        instance.robot_name = hrdf_path.relative_to(ROBOTS_PATH).as_posix()
        instance.body_parent_id = meta_info["body_parent_id"]
        instance.mesh_file_type = meta_info.get("mesh_file_type", None)
        instance.simulator_config = SimulatorConfig.from_dict(meta_info.get("simulator_config", {}))
        instance.body_name_config = BodyNameConfig.from_dict(meta_info.get("body_name_config", {}))
        instance.imu_config_list = [IMUConfig.from_dict(imu_config) for imu_config in meta_info.get("imu_config_list", [])]
        instance.motor_config_list = [MotorConfig.from_dict(motor_config) for motor_config in meta_info.get("motor_config_list", [])]

        if Path(hrdf_path, "joint_mapping.yaml").exists():
            instance.joint_mapping_config = JointMappingConfig.from_yaml(Path(hrdf_path, "joint_mapping.yaml"))

        for name in ["body", "joint", "actuator", "mesh", "simple_geom"]:
            component_csv = Path(hrdf_path, f"{name}.csv")
            if component_csv.exists():
                instance.info_list_dict[name] = InfoList.from_csv(str(component_csv), instance.info_class_dict[name])
        return instance

    @property
    def hrdf_path(self):
        return ROBOTS_PATH / self.robot_name
    
    def find_info_by_attr(self, attr_name: str, attr_value: str, info_name: str, single=False):
        assert info_name in self.info_class_dict, f"Info name {info_name} not found"
        return self.info_list_dict[info_name].find_info_by_attr(attr_name, attr_value, single)

    def save(self, mesh_path=None, max_faces=8000):
        """
        Save the HumanoidRobot to HRDF format.
        
        Args:
            mesh_path: Dictionary mapping mesh names to source paths (optional, for mesh processing)
            max_faces: Maximum number of faces for mesh simplification
        """
        self.hrdf_path.mkdir(parents=True, exist_ok=True)
        
        # Process mesh files if mesh_path is provided
        if mesh_path:
            for name, path in mesh_path.items():
                new_mesh_file = self.hrdf_path / "meshes" / f"{name}.{self.mesh_file_type}"
                new_mesh_file.parent.mkdir(parents=True, exist_ok=True)
                if path != new_mesh_file:
                    simplify_obj(path, new_mesh_file, max_faces)
        
        # Save metadata
        meta_path = self.hrdf_path / "meta.yaml"
        meta_path.touch(exist_ok=True)
        meta_info = {
            "body_parent_id": self.body_parent_id,
            "mesh_file_type": self.mesh_file_type,
            "simulator_config": self.simulator_config.to_dict(),
            "body_name_config": self.body_name_config.to_dict(),
            "imu_config_list": [imu_config.to_dict() for imu_config in self.imu_config_list],
            "motor_config_list": [motor_config.to_dict() for motor_config in self.motor_config_list],
        }
        with open(meta_path, "w", encoding='utf-8') as yaml_file:
            yaml.dump(meta_info, yaml_file, default_flow_style=False, allow_unicode=True, indent=2)
        
        # Save component CSV files
        for info_name in ["body", "joint", "actuator", "mesh", "simple_geom"]:
            if len(self.info_list_dict[info_name]) > 0:
                self.info_list_dict[info_name].save_csv(self.hrdf_path / f"{info_name}.csv")

    def fix_simple_geom(self):
        for idx, simple_geom in enumerate(self.info_list_dict["simple_geom"]):
            if simple_geom["name"].data == "":
                simple_geom["name"].data = f"geom_{idx}"

    def get_info_data_dict(self, info_name: str, key_attr: str, value_attr: str):
        assert info_name in self.info_class_dict, f"Info name {info_name} not found"
        return self.info_list_dict[info_name].get_data_dict(key_attr, value_attr)

    def get_info_data_list(self, info_name: str, attr: str):
        assert info_name in self.info_class_dict, f"Info name {info_name} not found"
        return self.info_list_dict[info_name].get_data_list(attr)

    # joint
    @property
    def joint_armature_dict(self):
        return self.get_info_data_dict("joint", "name", "armature")
    
    @property
    def joint_static_friction_dict(self):
        return self.get_info_data_dict("joint", "name", "static_friction")

    @property
    def joint_dynamic_friction_dict(self):
        return self.get_info_data_dict("joint", "name", "dynamic_friction")

    @property
    def joint_viscous_friction_dict(self):
        return self.get_info_data_dict("joint", "name", "viscous_friction")

    @property
    def joint_range_dict(self):
        return self.get_info_data_dict("joint", "name", "range")

    @property
    def joint_name_list(self):
        return self.get_info_data_list("joint", "name")

    # actuator
    @property
    def actuator_peak_velocity_dict(self):
        return self.get_info_data_dict("actuator", "joint_name", "peak_velocity")

    @property
    def actuator_peak_torque_dict(self):
        return self.get_info_data_dict("actuator", "joint_name", "peak_torque")

    @property
    def actuator_p_gain_dict(self):
        return self.get_info_data_dict("actuator", "joint_name", "p_gain")

    @property
    def actuator_d_gain_dict(self):
        return self.get_info_data_dict("actuator", "joint_name", "d_gain")
    
    @property
    def base_height(self):
        return float(self.info_list_dict["body"][0]["pos"].data[2])

    @property
    def hip_names(self):
        return self.body_name_config.hip_names

    @property
    def knee_names(self):
        return self.body_name_config.knee_names

    @property
    def foot_names(self):
        return self.body_name_config.foot_names

    @property
    def torso_name(self):
        return self.body_name_config.torso_name
    
    @property
    def imu_dict(self):
        return {imu_config.name: imu_config.value for imu_config in self.imu_config_list}

    @property
    def motor_dict(self):
        return {motor_config.name: motor_config.value for motor_config in self.motor_config_list}